"""
    Game Class which has bound multiple elements
    It also defines moves for the same
"""

from .gamemaze import GameMaze
from .pacman import Pacman
from .exceptions import KillGhostException
from .exceptions import KillPlayerException


class Game(object):
    """
        Game Class.
        Creates a new instance of this every resulting in a new game.
    """

    def __init__(self, difficulty=4, game_maze=GameMaze(randomMaze=True), score=0):
        object.__init__(self)
        self.difficulty = difficulty
        self.game_maze = game_maze
        self.ghosts = []
        self.pacman = Pacman(
            (game_maze.size_x / 2, game_maze.size_y * 3 / 4), score)

    def pacman_move_next(self, pacman):
        """
            Move next for Pacman
        """

        game_point = pacman.get_current()
        if not self.game_maze.checkwall(pacman.get_next()):
            pacman.move_next()

        if self.game_maze.get_attribute(pacman.get_current()) == GameMaze.MAZE['ghost'] or self.game_maze.get_attribute(pacman.get_current()) == GameMaze.MAZE['coinghost']:
            raise KillGhostException

        if self.game_maze.get_attribute(pacman.get_current()) == GameMaze.MAZE['coin']:
            pacman.change_score(1)
            self.game_maze.coin_count = self.game_maze.coin_count - 1

        self.game_maze.clear_all(game_point)

        self.game_maze.set_attribute(pacman.get_current(), GameMaze.MAZE['pacman'])

    def ghost_move_next(self, ghost):
        """
            Move next for ghost
        """

        game_point = ghost.get_current()
        ghost.think_next(self.game_maze, self.pacman)
        if not self.game_maze.checkwall(ghost.get_next()) and self.game_maze.get_attribute(ghost.get_next()) != GameMaze.MAZE['ghost'] and self.game_maze.get_attribute(ghost.get_next()) != GameMaze.MAZE['coinghost']:
            ghost.move_next()

        if self.game_maze.get_attribute(ghost.get_current()) == GameMaze.MAZE['pacman']:
            raise KillPlayerException

        if self.game_maze.get_attribute(game_point) == GameMaze.MAZE['coinghost']:
            self.game_maze.set_attribute(game_point, GameMaze.MAZE['coin'])
        else:
            self.game_maze.clear_all(game_point)

        if self.game_maze.get_attribute(ghost.get_current()) == GameMaze.MAZE['coin']:
            self.game_maze.set_attribute(
                ghost.get_current(), GameMaze.MAZE['coinghost'])
        else:
            self.game_maze.set_attribute(
                ghost.get_current(), GameMaze.MAZE['ghost'])

    def multiple_move_next(self, thing):
        """ Move Next for all objects """

        if thing.__class__.__name__ == 'Pacman':
            self.pacman_move_next(thing)
        elif thing.__class__.__name__ == 'Ghost':
            self.ghost_move_next(thing)
        else:
            print 'Whoa Dude! How Do You Even?'

    def check_ghost(self):
        """
            Check if ghost on spot
        """

        return self.game_maze.get_attribute(self.pacman.get_current()) == GameMaze.MAZE['ghost']



"""
    A GameMaze object which contains helpers for handling mazes
"""

from random import randint


class GameMaze(object):
    """
        Making A Maze Module For Game
    """

    MAZE = dict(
        wall='X',
        ghost='G',
        pacman='P',
        coin='C',
        nothing='.',
        coinghost='CG',
    )
    COLORS = dict(
        wall='\033[95m',
        ghost='\033[37m',
        coin='\033[33m',
        pacman='\033[91m',
        coinghost='\033[47m',
        nothing='\033[36m',
        END='\033[0m',
    )

    def __init__(self, randomMaze=False, size=(35, 15)):
        object.__init__(self)
        self.size_x = size[0]
        self.size_y = size[1]
        self.coin_count = 0
        self.game_maze = [x[:] for x in [['.'] * self.size_x] * self.size_y]
        self.free_space = set()

        for i in xrange(0, self.size_y):
            for j in [0, self.size_x - 1]:
                self.game_maze[i][j] = GameMaze.MAZE['wall']
        for i in xrange(0, self.size_x):
            for j in [0, self.size_y - 1]:
                self.game_maze[j][i] = GameMaze.MAZE['wall']

        if randomMaze is False:
            pass
        else:
            self.fill_wall()

        self.game_maze[self.size_y * 3 / 4][self.size_x / 2] = GameMaze.MAZE['pacman']
        self.fill_coin([x[:] for x in [[0] * self.size_x] * self.size_y])

    def place_enemy(self):
        """
            Places enemy on map
        """

        size_x = randint(1, self.size_x - 1)
        size_y = randint(1, self.size_y - 1)
        while self.checkwall((size_x, size_y)) is True and self.get_attribute((size_x, size_y)) != GameMaze.MAZE['coinghost'] and self.get_attribute((size_x, size_y)) != GameMaze.MAZE['ghost'] and self.get_attribute((size_x, size_y)) != GameMaze.MAZE['coin'] and (size_x, size_y) not in self.free_space:
            size_x = randint(1, self.size_x - 1)
            size_y = randint(1, self.size_y - 1)
        attribute = self.get_attribute((size_x, size_y))
        if attribute == GameMaze.MAZE['coin']:
            self.set_attribute((size_x, size_y), GameMaze.MAZE['coinghost'])
        else:
            self.set_attribute((size_x, size_y), GameMaze.MAZE['ghost'])
        return (size_x, size_y)

    def fill_wall(self, game_point=(2, 2), count=0.0):
        """
            Fills Wall in map
        """

        point_1 = list(game_point)
        point_2 = [game_point[0], self.size_y - game_point[1] - 1]
        point_3 = [self.size_x - game_point[0] - 1, game_point[1]]
        point_4 = [self.size_x - game_point[0] - 1, self.size_y - game_point[1] - 1]

        total = self.size_x * self.size_y

        while point_1[0] < self.size_x / 2 and point_1[1] < self.size_y / 2:
            count += 4.0
            self.make_wall(point_1)
            self.make_wall(point_2)
            self.make_wall(point_3)
            self.make_wall(point_4)
            if randint(0, 1) == 0:
                point_1[0] += 1
                point_2[0] += 1
                point_3[0] -= 1
                point_4[0] -= 1
            else:
                point_1[1] += 1
                point_2[1] -= 1
                point_3[1] += 1
                point_4[1] -= 1
        if count / total < 0.3:
            self.fill_wall((randint(3, self.size_x - 3), randint(3, self.size_x - 3)), count)

    def fill_coin(self, visited, game_point=(1, 1)):
        """
            Fills coins on map
        """

        if self.checkwall(game_point) is True or visited[game_point[1]][game_point[0]] is True or self.get_attribute(game_point) == GameMaze.MAZE['pacman']:
            return
        self.free_space.add(game_point)
        visited[game_point[1]][game_point[0]] = True
        if randint(0, 2) == 0:  # Probhablisticly fill with coin
            self.make_coin(game_point)
            self.coin_count = self.coin_count + 1
        self.fill_coin(visited, (game_point[0], game_point[1] + 1))
        self.fill_coin(visited, (game_point[0] + 1, game_point[1]))
        self.fill_coin(visited, (game_point[0], game_point[1] - 1))
        self.fill_coin(visited, (game_point[0] - 1, game_point[1]))

    def set_attribute(self, game_point, attribute):
        """
            Sets attribute on point
        """

        self.game_maze[game_point[1]][game_point[0]] = attribute

    def get_attribute(self, game_point):
        """
            Gets attribute on point
        """

        return self.game_maze[game_point[1]][game_point[0]]

    def checkwall(self, game_point):
        """
            Checks Wall
        """

        return self.get_attribute(game_point) == GameMaze.MAZE['wall']

    def clear_all(self, game_point):
        """
            Clears all on point
        """

        self.set_attribute(game_point, GameMaze.MAZE['nothing'])

    def make_wall(self, game_point):
        """
            Makes Wall
        """

        self.set_attribute(game_point, GameMaze.MAZE['wall'])

    def make_coin(self, game_point):
        """
            Makes Coin on point
        """

        self.set_attribute(game_point, GameMaze.MAZE['coin'])

    def __str__(self):
        string = ''
        for i in self.game_maze:
            for j in i:
                if j == GameMaze.MAZE['wall']:
                    string += GameMaze.COLORS['wall'] + j + GameMaze.COLORS['END'] + ' '
                elif j == GameMaze.MAZE['coin']:
                    string += GameMaze.COLORS['coin'] + j + GameMaze.COLORS['END'] + ' '
                elif j == GameMaze.MAZE['pacman']:
                    string += GameMaze.COLORS['pacman'] + j + GameMaze.COLORS['END'] + ' '
                elif j == GameMaze.MAZE['ghost']:
                    string += GameMaze.COLORS['ghost'] + j + GameMaze.COLORS['END'] + ' '
                elif j == GameMaze.MAZE['coinghost']:
                    string += GameMaze.COLORS['coinghost'] + GameMaze.MAZE['ghost'] + GameMaze.COLORS['END'] + ' '
                elif j == GameMaze.MAZE['nothing']:
                    string += GameMaze.COLORS['nothing'] + j + GameMaze.COLORS['END'] + ' '
                else:
                    string += j + ' '
            string += '\n'
        return string

    __repr__ = __str__



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
    Pacman implementation in python
"""

from .game import Game
from .gamemaze import GameMaze
from .ghost import Ghost
from .utils import clear, argument_parser, getch


def main():
    """
        Main function
    """

    (size_x, size_y, difficulty) = argument_parser()
    game_maze = GameMaze(1, (size_x, size_y))
    game = Game(difficulty, game_maze)
    for _ in xrange(0, difficulty):
        game.ghosts.append(Ghost(game.game_maze.place_enemy()))
    character = getch()
    while True:
        clear()  # default would be enough
        game.multiple_move_next(game.pacman)
        for i in game.ghosts:
            game.multiple_move_next(i)
        print 'Score : ' + str(game.pacman.score)
        print game.game_maze
        character = getch()
        while character not in ['q', 'w', 'a', 's', 'd']:
            character = getch()
        if character == 'q':
            exit(0)
        elif character == 'w':
            game.pacman.move_up()
        elif character == 'a':
            game.pacman.move_left()
        elif character == 's':
            game.pacman.move_down()
        elif character == 'd':
            game.pacman.move_right()
        if game.game_maze.coin_count == 0:
            print 'LEVEL WON! BOND!!!'
            character = getch()
            game_maze = GameMaze(1, (size_x, size_y))
            game = Game(difficulty, game_maze, game.pacman.score)
            for i in xrange(0, difficulty):
                game.ghosts.append(Ghost(game.game_maze.place_enemy()))


if __name__ == '__main__':
    main()



"""
    A generic class for Persons
"""


class Person(object):
    """
        Generic object
    """

    def __init__(self, position):
        object.__init__(self)
        self.position = list(position)
        self.direction_vector = (0, 0)

    def move_up(self):
        """
            Moves move_up
        """

        self.direction_vector = (0, -1)

    def move_down(self):
        """
            Moves move_down
        """

        self.direction_vector = (0, 1)

    def move_left(self):
        """
            Moves left
        """

        self.direction_vector = (-1, 0)

    def move_right(self):
        """
            Moves move_right
        """

        self.direction_vector = (1, 0)

    def get_current(self):
        """
            Gets Current position
        """

        return self.position[::]

    def move_next(self):
        """
            Moves Next position
        """

        self.position[0] += self.direction_vector[0]
        self.position[1] += self.direction_vector[1]

    def get_next(self):
        """
            Gets tentative next position
        """

        return (self.position[0] + self.direction_vector[0], self.position[1] + self.direction_vector[1])



"""
    A collection of helper modules
"""

import argparse

try:
    from msvcrt import getch
except ImportError:
    from sys import stdin
    from termios import tcgetattr, tcsetattr, TCSADRAIN
    from tty import setraw

    def getch():
        """
            GetCh wrapper for linux
        """

        file_descriptor = stdin.fileno()
        old_settings = tcgetattr(file_descriptor)
        try:
            setraw(file_descriptor)
            character = stdin.read(1)
        finally:
            tcsetattr(file_descriptor, TCSADRAIN, old_settings)
        return character


def clear(clear_num=100):
    """
        clears the screen by printing too many blank lines
    """
    for _ in xrange(clear_num):
        print


def size_parser():
    """
        DEPRICATED: Parse Sizes
    """

    try:
        size_x = int(raw_input('Enter The Size-X (odd number)\n>'))
        size_y = int(raw_input('Enter The Size-Y (odd number)\n>'))
        if size_x < 8 or size_y < 8:
            print "Not that small dude ('-_-) "
            raise ValueError()
        if size_x > 80 or size_y > 40:
            print 'Size too much'
            raise ValueError()
        if size_x % 2 == 0:
            size_x = size_x + 1
        if size_y % 2 == 0:
            size_y = size_y + 1
    except ValueError:
        print 'Invalid Input. Assuming Defaults (35,15)'
        (size_x, size_y) = (35, 15)
    return (size_x, size_y)


def difficulty_parser():
    """
        DEPRICATED: Parse difficulty
    """

    try:
        difficulty = int(
            raw_input('Enter the difficulty (Number of Ghosts)\n>'))
        if difficulty < 0:
            print "Not that small dude ('-_-) "
            raise ValueError()
        if difficulty > 10:
            print 'Too many ghosts'
            raise ValueError()
    except ValueError:
        print 'Invalid Input. Assuming Defaults 4 Ghosts'
        difficulty = 4
    return difficulty


def argument_parser():
    """
        Parse arguments from the command line
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--legacy", action='store_true')
    parser.add_argument('-d', '--difficulty', type=int, help="Enter the difficulty [0-10]", default=4)
    parser.add_argument('-x', '--size-x', type=int, help="Enter the width", default=35)
    parser.add_argument('-y', '--size-y', type=int, help="Enter the height", default=15)
    args = parser.parse_args()

    if args.legacy:
        size_x, size_y = size_parser()
        difficulty = difficulty_parser()
        return (size_x, size_y, difficulty)

    return (args.size_x, args.size_y, args.difficulty)



"""
    A set of custom exceptions defined for pacman game
"""


class KillGhostException(Exception):
    """
        A exception when player attempts to kill a ghost
    """
    def __init__(self):
        Exception.__init__(self)


class KillPlayerException(Exception):
    """
        An exception raised when ghost sucessfully kills a player
    """
    def __init__(self):
        Exception.__init__(self)



"""
    A pacman object instance
"""

from .person import Person


class Pacman(Person):
    """
        Generic Pacman Object
    """

    def __init__(self, position, score):
        Person.__init__(self, position)
        self.score = score

    def change_score(self, number):
        """
            changes score
        """

        self.score += number

    def collect_coin(self, score):
        """
            Duplicate of change_score
        """
        self.change_score(score)



from setuptools import setup, find_packages
from codecs import open
from os import path

VERSION = '1.0.2.dev1'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if 'git+' not in x]

setup(
    name='pacman-game',
    version=VERSION,
    description='A Pacman Like Game with new and interesting new features.',
    long_description=long_description,
    url='https://github.com/ParthKolekar/pacman-game',
    download_url='https://github.com/ParthKolekar/pacman-game/tarball/' + VERSION,
    license='BSD',
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 2',
    ],
    keywords='pacman console ai game',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author='Parth Kolekar',
    install_requires=install_requires,
    depedency_links=dependency_links,
    author_email='parth.kolekar@students.iiit.ac.in',
    entry_points={
        'console_scripts': [
            'pacman-game=pacman_game.main:main',
        ],
    },
)



"""
    A smart AI based Ghost
"""

from random import randint

from .person import Person


class Ghost(Person):
    """
        Smart AI based Ghosts
    """

    def __init__(self, position, intelligence=0):
        Person.__init__(self, position)
        self.intelligence = intelligence

    def move_towards_pacman_greedy(self, pacman):
        """
            Greedily move towards pacman.
            Might get stuck. (Intendended for dumb AI)
        """
        pass  # TODO

    def move_towards_pacman_smart(self, pacman, game_maze):
        """
            Uses A* search to get best move towards pacman
        """
        pass  # TODO

    def move_towards_pacman_random(self):
        """
            Randombly moves anywhere.
        """
        random_number = randint(0, 4)  # 1/5 to maintain too
        if random_number == 0:
            self.move_up()
        elif random_number == 1:
            self.move_down()
        elif random_number == 2:
            self.move_left()
        elif random_number == 3:
            self.move_right()

    def think_next(self, game_maze, pacman):
        """
            Switches between various methods of finding next moves.
            Based on intelligence
        """
        if self.intelligence == 0:
            self.move_towards_pacman_random()
        else:
            pass  # TODO
